@@ -2,7 +2,7 @@
  * ccache -- a fast C/C++ compiler cache
  *
  * Copyright (C) 2002-2007 Andrew Tridgell
- * Copyright (C) 2009-2012 Joel Rosdahl
+ * Copyright (C) 2009-2016 Joel Rosdahl
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -36,7 +36,7 @@ static const char VERSION_TEXT[] =
 MYNAME " version %s\n"
 "\n"
 "Copyright (C) 2002-2007 Andrew Tridgell\n"
-"Copyright (C) 2009-2011 Joel Rosdahl\n"
+"Copyright (C) 2009-2016 Joel Rosdahl\n"
 "\n"
 "This program is free software; you can redistribute it and/or modify it under\n"
 "the terms of the GNU General Public License as published by the Free Software\n"
@@ -197,6 +197,14 @@ enum fromcache_call_mode {
 	FROMCACHE_COMPILED_MODE
 };
 
+struct pending_tmp_file {
+	char *path;
+	struct pending_tmp_file *next;
+};
+
+/* Temporary files to remove at program exit. */
+static struct pending_tmp_file *pending_tmp_files = NULL;
+
 /*
  * This is a string that identifies the current "version" of the hash sum
  * computed by ccache. If, for any reason, we want to force the hash sum to be
@@ -231,23 +239,87 @@ failed(void)
 }
 
 static void
-clean_up_tmp_files()
+add_pending_tmp_file(const char *path)
 {
-	/* delete intermediate pre-processor file if needed */
-	if (i_tmpfile) {
-		if (!direct_i_file) {
-			tmp_unlink(i_tmpfile);
-		}
-		free(i_tmpfile);
-		i_tmpfile = NULL;
+	struct pending_tmp_file *e = x_malloc(sizeof(*e));
+	e->path = x_strdup(path);
+	e->next = pending_tmp_files;
+	pending_tmp_files = e;
+}
+
+static void
+clean_up_pending_tmp_files(void)
+{
+	struct pending_tmp_file *p = pending_tmp_files;
+	while (p) {
+		tmp_unlink(p->path);
+		p = p->next;
+		/* Leak p->path and p here because clean_up_pending_tmp_files needs to be signal
+		 * safe. */
+	}
+}
+
+static void
+signal_handler(int signo)
+{
+	(void)signo;
+	clean_up_pending_tmp_files();
+}
+
+static void
+clean_up_internal_tempdir(void)
+{
+	DIR *dir;
+	struct dirent *entry;
+	struct stat st;
+	time_t now = time(NULL);
+
+	stat(cache_dir, &st);
+	if (st.st_mtime + 3600 >= now) {
+		/* No cleanup needed. */
+		return;
 	}
 
-	/* delete the cpp stderr file if necessary */
-	if (cpp_stderr) {
-		tmp_unlink(cpp_stderr);
-		free(cpp_stderr);
-		cpp_stderr = NULL;
+	update_mtime(cache_dir);
+
+	dir = opendir(temp_dir);
+	if (!dir) {
+		return;
 	}
+
+	while ((entry = readdir(dir))) {
+		char *path;
+
+		if (str_eq(entry->d_name, ".") || str_eq(entry->d_name, "..")) {
+			continue;
+		}
+
+		path = format("%s/%s", temp_dir, entry->d_name);
+		if (lstat(path, &st) == 0 && st.st_mtime + 3600 < now) {
+			tmp_unlink(path);
+		}
+		free(path);
+	}
+
+	closedir(dir);
+}
+
+static char *
+get_current_working_dir(void)
+{
+	if (!current_working_dir) {
+		char *cwd = get_cwd();
+		if (cwd) {
+			current_working_dir = x_realpath(cwd);
+			free(cwd);
+		}
+		if (!current_working_dir) {
+			cc_log("Unable to determine current working directory: %s",
+			       strerror(errno));
+			failed();
+		}
+	}
+	return current_working_dir;
 }
 
 /*
@@ -266,7 +338,7 @@ get_path_in_cache(const char *name, const char *suffix)
 		char *p = format("%s/%c", path, name[i]);
 		free(path);
 		path = p;
-		if (create_dir(path) != 0) {
+		if (!getenv("CCACHE_READONLY") && create_dir(path) != 0) {
 			fatal("Failed to create %s: %s", path, strerror(errno));
 		}
 	}
@@ -380,21 +452,57 @@ ignore:
 }
 
 /*
- * Make a relative path from CCACHE_BASEDIR to path. Takes over ownership of
- * path. Caller frees.
+ * Make a relative path from current working directory to path if
+ * CCACHE_BASEDIR is a prefix of path. Takes over ownership of path. Caller
+ * frees.
  */
 static char *
 make_relative_path(char *path)
 {
-	char *relpath;
+	char *relpath, *canon_path, *path_suffix = NULL;
+	struct stat st;
 
 	if (!base_dir || !str_startswith(path, base_dir)) {
 		return path;
 	}
 
-	relpath = get_relative_path(current_working_dir, path);
-	free(path);
-	return relpath;
+	/* x_realpath only works for existing paths, so if path doesn't exist, try
+	 * dirname(path) and assemble the path afterwards. We only bother to try
+	 * canonicalizing one of these two paths since a compiler path argument
+	 * typically only makes sense if path or dirname(path) exists. */
+	if (stat(path, &st) != 0) {
+		/* path doesn't exist. */
+		char *dir, *p;
+		dir = dirname(path);
+		if (stat(dir, &st) != 0) {
+			/* And neither does its parent directory, so no action to take. */
+			free(dir);
+			return path;
+		}
+		path_suffix = basename(path);
+		p = path;
+		path = dirname(path);
+		free(p);
+	}
+
+	canon_path = x_realpath(path);
+	if (canon_path) {
+		free(path);
+		relpath = get_relative_path(get_current_working_dir(), canon_path);
+		free(canon_path);
+		if (path_suffix) {
+			path = format("%s/%s", relpath, path_suffix);
+			free(relpath);
+			free(path_suffix);
+			return path;
+		} else {
+			return relpath;
+		}
+	} else {
+		/* path doesn't exist, so leave it as it is. */
+		free(path_suffix);
+		return path;
+	}
 }
 
 /*
@@ -530,7 +638,13 @@ to_cache(struct args *args)
 	args_pop(args, 3);
 
 	if (stat(tmp_stdout, &st) != 0) {
-		fatal("Could not create %s (permission denied?)", tmp_stdout);
+		/* The stdout file was removed - cleanup in progress? Better bail out. */
+		cc_log("%s not found: %s", tmp_stdout, strerror(errno));
+		stats_update(STATS_MISSING);
+		tmp_unlink(tmp_stdout);
+		tmp_unlink(tmp_stderr);
+		tmp_unlink(tmp_obj);
+		failed();
 	}
 	if (st.st_size != 0) {
 		cc_log("Compiler produced stdout");
@@ -693,14 +807,16 @@ get_object_name_from_cpp(struct args *args, struct mdfour *hash)
 		input_base[10] = 0;
 	}
 
-	/* now the run */
-	path_stdout = format("%s/%s.tmp.%s.%s",
-	                     temp_dir, input_base, tmp_string(), i_extension);
 	path_stderr = format("%s/tmp.cpp_stderr.%s", temp_dir, tmp_string());
+	add_pending_tmp_file(path_stderr);
 
 	time_of_compilation = time(NULL);
 
 	if (!direct_i_file) {
+		path_stdout = format("%s/%s.tmp.%s.%s",
+		                     temp_dir, input_base, tmp_string(), i_extension);
+		add_pending_tmp_file(path_stdout);
+
 		/* run cpp on the input file to obtain the .i */
 		args_add(args, "-E");
 		args_add(args, input_file);
@@ -720,10 +836,6 @@ get_object_name_from_cpp(struct args *args, struct mdfour *hash)
 	}
 
 	if (status != 0) {
-		if (!direct_i_file) {
-			tmp_unlink(path_stdout);
-		}
-		tmp_unlink(path_stderr);
 		cc_log("Preprocessor gave exit status %d", status);
 		stats_update(STATS_PREPROCESSOR);
 		failed();
@@ -740,7 +852,6 @@ get_object_name_from_cpp(struct args *args, struct mdfour *hash)
 		hash_delimiter(hash, "unifycpp");
 		if (unify_hash(hash, path_stdout) != 0) {
 			stats_update(STATS_ERROR);
-			tmp_unlink(path_stderr);
 			cc_log("Failed to unify %s", path_stdout);
 			failed();
 		}
@@ -748,7 +859,6 @@ get_object_name_from_cpp(struct args *args, struct mdfour *hash)
 		hash_delimiter(hash, "cpp");
 		if (!process_preprocessed_file(hash, path_stdout)) {
 			stats_update(STATS_ERROR);
-			tmp_unlink(path_stderr);
 			failed();
 		}
 	}
@@ -768,7 +878,8 @@ get_object_name_from_cpp(struct args *args, struct mdfour *hash)
 		 */
 		cpp_stderr = path_stderr;
 	} else {
-		tmp_unlink(path_stderr);
+		hash_delimiter(hash, "runsecondcpp");
+		hash_string(hash, "false");
 		free(path_stderr);
 	}
 
@@ -915,6 +1026,13 @@ calculate_object_hash(struct args *args, struct mdfour *hash, int direct_mode)
 			continue;
 		}
 
+		/* The -fdebug-prefix-map option may be used in combination with
+		   CCACHE_BASEDIR to reuse results across different directories. Skip it
+		   from hashing. */
+		if (str_startswith(args->argv[i], "-fdebug-prefix-map=")) {
+			continue;
+		}
+
 		/* When using the preprocessor, some arguments don't contribute
 		   to the hash. The theory is that these arguments will change
 		   the output of -E if they are going to have any effect at
@@ -951,6 +1069,16 @@ calculate_object_hash(struct args *args, struct mdfour *hash, int direct_mode)
 			continue;
 		}
 
+		if (str_eq(args->argv[i], "-Xclang")
+		    && i + 3 < args->argc
+		    && str_eq(args->argv[i+1], "-load")
+		    && str_eq(args->argv[i+2], "-Xclang")
+		    && stat(args->argv[i+3], &st) == 0) {
+			hash_delimiter(hash, "plugin");
+			hash_compiler(hash, &st, args->argv[i+3], false);
+			continue;
+		}
+
 		/* All other arguments are included in the hash. */
 		hash_delimiter(hash, "arg");
 		hash_string(hash, args->argv[i]);
@@ -1039,6 +1167,16 @@ from_cache(enum fromcache_call_mode mode, bool put_object_in_manifest)
 		return;
 	}
 
+	/*
+	 * Occasionally, e.g. on hard reset, our cache ends up as just filesystem
+	 * meta-data with no content catch an easy case of this.
+	 */
+	if (st.st_size == 0) {
+		cc_log("Invalid (empty) object file %s in cache", cached_obj);
+		x_unlink(cached_obj);
+		return;
+	}
+
 	/*
 	 * (If mode != FROMCACHE_DIRECT_MODE, the dependency file is created by
 	 * gcc.)
@@ -1170,12 +1308,12 @@ from_cache(enum fromcache_call_mode mode, bool put_object_in_manifest)
 	/* log the cache hit */
 	switch (mode) {
 	case FROMCACHE_DIRECT_MODE:
-		cc_log("Succeded getting cached result");
+		cc_log("Succeeded getting cached result");
 		stats_update(STATS_CACHEHIT_DIR);
 		break;
 
 	case FROMCACHE_CPP_MODE:
-		cc_log("Succeded getting cached result");
+		cc_log("Succeeded getting cached result");
 		stats_update(STATS_CACHEHIT_CPP);
 		break;
 
@@ -1371,14 +1509,14 @@ cc_process_args(struct args *orig_args, struct args **preprocessor_args,
 				result = false;
 				goto out;
 			}
-			output_obj = argv[i+1];
+			output_obj = make_relative_path(x_strdup(argv[i+1]));
 			i++;
 			continue;
 		}
 
 		/* alternate form of -o, with no space */
 		if (str_startswith(argv[i], "-o")) {
-			output_obj = &argv[i][2];
+			output_obj = make_relative_path(x_strdup(&argv[i][2]));
 			continue;
 		}
 
@@ -1412,10 +1550,10 @@ cc_process_args(struct args *orig_args, struct args **preprocessor_args,
 		}
 		if (str_startswith(argv[i], "-MF")) {
 			char *arg;
+			bool separate_argument = (strlen(argv[i]) == 3);
 			dependency_filename_specified = true;
 			free(output_dep);
-			args_add(dep_args, argv[i]);
-			if (strlen(argv[i]) == 3) {
+			if (separate_argument) {
 				/* -MF arg */
 				if (i >= argc - 1) {
 					cc_log("Missing argument to %s", argv[i]);
@@ -1424,18 +1562,26 @@ cc_process_args(struct args *orig_args, struct args **preprocessor_args,
 					goto out;
 				}
 				arg = argv[i + 1];
-				args_add(dep_args, argv[i + 1]);
 				i++;
 			} else {
 				/* -MFarg */
 				arg = &argv[i][3];
 			}
 			output_dep = make_relative_path(x_strdup(arg));
+			/* Keep the format of the args the same */
+			if (separate_argument) {
+				args_add(dep_args, "-MF");
+				args_add(dep_args, output_dep);
+			} else {
+				char *option = format("-MF%s", output_dep);
+				args_add(dep_args, option);
+				free(option);
+			}
 			continue;
 		}
 		if (str_startswith(argv[i], "-MQ") || str_startswith(argv[i], "-MT")) {
+			char *relpath;
 			dependency_target_specified = true;
-			args_add(dep_args, argv[i]);
 			if (strlen(argv[i]) == 3) {
 				/* -MQ arg or -MT arg */
 				if (i >= argc - 1) {
@@ -1444,8 +1590,21 @@ cc_process_args(struct args *orig_args, struct args **preprocessor_args,
 					result = false;
 					goto out;
 				}
-				args_add(dep_args, argv[i + 1]);
+				args_add(dep_args, argv[i]);
+				relpath = make_relative_path(x_strdup(argv[i + 1]));
+				args_add(dep_args, relpath);
+				free(relpath);
 				i++;
+			} else {
+				char *arg_opt;
+				char *option;
+				arg_opt = x_strndup(argv[i], 3);
+				relpath = make_relative_path(x_strdup(argv[i] + 3));
+				option = format("%s%s", arg_opt, relpath);
+				args_add(dep_args, option);
+				free(arg_opt);
+				free(relpath);
+				free(option);
 			}
 			continue;
 		}
@@ -1588,8 +1747,17 @@ cc_process_args(struct args *orig_args, struct args **preprocessor_args,
 			goto out;
 		}
 
-		/* Rewrite to relative to increase hit rate. */
-		input_file = make_relative_path(x_strdup(argv[i]));
+		lstat(argv[i], &st);
+		if (S_ISLNK(st.st_mode)) {
+			/* Don't rewrite source file path if it's a symlink since
+			   make_relative_path resolves symlinks using realpath(3) and this leads
+			   to potentially choosing incorrect relative header files. See the
+			   "symlink to source file" test. */
+			input_file = x_strdup(argv[i]);
+		} else {
+			/* Rewrite to relative to increase hit rate. */
+			input_file = make_relative_path(x_strdup(argv[i]));
+		}
 	}
 
 	if (!input_file) {
@@ -1602,8 +1770,8 @@ cc_process_args(struct args *orig_args, struct args **preprocessor_args,
 	if (found_pch || found_fpch_preprocess) {
 		using_precompiled_header = true;
 		if (!(sloppiness & SLOPPY_TIME_MACROS)) {
-			cc_log("You have to specify \"time_macros\" sloppiness when using"
-			       " precompiled headers to get direct hits");
+			cc_log("You have to specify \"pch_defines,time_macros\" sloppiness when"
+			       " using precompiled headers to get direct hits");
 			cc_log("Disabling direct mode");
 			stats_update(STATS_CANTUSEPCH);
 			result = false;
@@ -1630,19 +1798,31 @@ cc_process_args(struct args *orig_args, struct args **preprocessor_args,
 	output_is_precompiled_header =
 		actual_language && strstr(actual_language, "-header") != NULL;
 
-	if (!found_c_opt && !output_is_precompiled_header) {
-		cc_log("No -c option found");
-		/* I find that having a separate statistic for autoconf tests is useful,
-		   as they are the dominant form of "called for link" in many cases */
-		if (strstr(input_file, "conftest.")) {
-			stats_update(STATS_CONFTEST);
-		} else {
-			stats_update(STATS_LINK);
-		}
+	if (output_is_precompiled_header && !(sloppiness & SLOPPY_PCH_DEFINES)) {
+		cc_log("You have to specify \"pch_defines,time_macros\" sloppiness when"
+		       " creating precompiled headers");
+		stats_update(STATS_CANTUSEPCH);
 		result = false;
 		goto out;
 	}
 
+	if (!found_c_opt) {
+		if (output_is_precompiled_header) {
+			args_add(stripped_args, "-c");
+		} else {
+			cc_log("No -c option found");
+			/* I find that having a separate statistic for autoconf tests is useful,
+			   as they are the dominant form of "called for link" in many cases */
+			if (strstr(input_file, "conftest.")) {
+				stats_update(STATS_CONFTEST);
+			} else {
+				stats_update(STATS_LINK);
+			}
+			result = false;
+			goto out;
+		}
+	}
+
 	if (!actual_language) {
 		cc_log("Unsupported source extension: %s", input_file);
 		stats_update(STATS_SOURCELANG);
@@ -1782,7 +1962,7 @@ cc_reset(void)
 	base_dir = NULL;
 	args_free(orig_args); orig_args = NULL;
 	free(input_file); input_file = NULL;
-	output_obj = NULL;
+	free(output_obj); output_obj = NULL;
 	free(output_dep); output_dep = NULL;
 	free(cached_obj_hash); cached_obj_hash = NULL;
 	free(cached_obj); cached_obj = NULL;
@@ -1832,15 +2012,42 @@ parse_sloppiness(char *p)
 			cc_log("Being sloppy about __DATE__ and __TIME__");
 			result |= SLOPPY_TIME_MACROS;
 		}
+		if (str_eq(word, "pch_defines")) {
+			cc_log("Being sloppy about PCH #defines");
+			result |= SLOPPY_PCH_DEFINES;
+		}
 		q = NULL;
 	}
 	free(p);
 	return result;
 }
 
+/* Make a copy of stderr that will not be cached, so things like
+   distcc can send networking errors to it. */
+static void
+setup_uncached_err(void)
+{
+	char *buf;
+	int uncached_fd;
+
+	uncached_fd = dup(2);
+	if (uncached_fd == -1) {
+		cc_log("dup(2) failed: %s", strerror(errno));
+		failed();
+	}
+
+	/* leak a pointer to the environment */
+	buf = format("UNCACHED_ERR_FD=%d", uncached_fd);
+
+	if (putenv(buf) == -1) {
+		cc_log("putenv failed: %s", strerror(errno));
+		failed();
+	}
+}
+
 /* the main ccache driver function */
 static void
-ccache(int argc, char *argv[])
+ccache(char *argv[])
 {
 	bool put_object_in_manifest = false;
 	struct file_hash *object_hash;
@@ -1856,12 +2063,7 @@ ccache(int argc, char *argv[])
 	/* Arguments to send to the real compiler. */
 	struct args *compiler_args;
 
-	find_compiler(argc, argv);
-
-	if (getenv("CCACHE_DISABLE")) {
-		cc_log("ccache is disabled");
-		failed();
-	}
+	setup_uncached_err();
 
 	if (!getenv("CCACHE_READONLY")) {
 		if (create_cachedirtag(cache_dir) != 0) {
@@ -1875,7 +2077,7 @@ ccache(int argc, char *argv[])
 
 	cc_log_argv("Command line: ", argv);
 	cc_log("Hostname: %s", get_hostname());
-	cc_log("Working directory: %s", current_working_dir);
+	cc_log("Working directory: %s", get_current_working_dir());
 
 	if (base_dir) {
 		cc_log("Base directory: %s", base_dir);
@@ -2111,39 +2313,20 @@ ccache_main_options(int argc, char *argv[])
 	return 0;
 }
 
-
-/* Make a copy of stderr that will not be cached, so things like
-   distcc can send networking errors to it. */
-static void
-setup_uncached_err(void)
-{
-	char *buf;
-	int uncached_fd;
-
-	uncached_fd = dup(2);
-	if (uncached_fd == -1) {
-		cc_log("dup(2) failed: %s", strerror(errno));
-		failed();
-	}
-
-	/* leak a pointer to the environment */
-	buf = format("UNCACHED_ERR_FD=%d", uncached_fd);
-
-	if (putenv(buf) == -1) {
-		cc_log("putenv failed: %s", strerror(errno));
-		failed();
-	}
-}
-
 int
 ccache_main(int argc, char *argv[])
 {
 	char *p;
 	char *program_name;
+	bool external_tempdir;
+
+	signal(SIGHUP, signal_handler);
+	signal(SIGINT, signal_handler);
+	signal(SIGTERM, signal_handler);
 
 	exitfn_init();
 	exitfn_add_nullary(stats_flush);
-	exitfn_add_nullary(clean_up_tmp_files);
+	exitfn_add_nullary(clean_up_pending_tmp_files);
 
 	/* check for logging early so cc_log messages start working ASAP */
 	cache_logfile = getenv("CCACHE_LOGFILE");
@@ -2160,11 +2343,6 @@ ccache_main(int argc, char *argv[])
 		}
 	}
 
-	current_working_dir = get_cwd();
-	if (!current_working_dir) {
-		cc_log("Could not determine current working directory");
-		failed();
-	}
 	cache_dir = getenv("CCACHE_DIR");
 	if (cache_dir) {
 		cache_dir = x_strdup(cache_dir);
@@ -2190,13 +2368,16 @@ ccache_main(int argc, char *argv[])
 	}
 	free(program_name);
 
-	check_cache_dir();
+	/* find_compiler sets up orig_args, needed by failed(), so do it early. */
+	find_compiler(argc, argv);
 
-	temp_dir = getenv("CCACHE_TEMPDIR");
-	if (!temp_dir) {
-		temp_dir = format("%s/tmp", cache_dir);
+	if (getenv("CCACHE_DISABLE")) {
+		cc_log("ccache is disabled");
+		failed();
 	}
 
+	check_cache_dir();
+
 	base_dir = getenv("CCACHE_BASEDIR");
 	if (base_dir && base_dir[0] != '/') {
 		cc_log("Ignoring non-absolute base directory %s", base_dir);
@@ -2205,8 +2386,6 @@ ccache_main(int argc, char *argv[])
 
 	compile_preprocessed_source_code = !getenv("CCACHE_CPP2");
 
-	setup_uncached_err();
-
 	/* make sure the cache dir exists */
 	if (create_dir(cache_dir) != 0) {
 		fprintf(stderr,
@@ -2216,6 +2395,13 @@ ccache_main(int argc, char *argv[])
 	}
 
 	/* make sure the temp dir exists */
+	temp_dir = getenv("CCACHE_TEMPDIR");
+	if (temp_dir) {
+		external_tempdir = true;
+	} else {
+		temp_dir = format("%s/tmp", cache_dir);
+		external_tempdir = false;
+	}
 	if (create_dir(temp_dir) != 0) {
 		fprintf(stderr,
 		        "ccache: failed to create %s (%s)\n",
@@ -2223,6 +2409,10 @@ ccache_main(int argc, char *argv[])
 		exit(1);
 	}
 
-	ccache(argc, argv);
+	if (!external_tempdir) {
+		clean_up_internal_tempdir();
+	}
+
+	ccache(argv);
 	return 1;
 }
