@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2002 Andrew Tridgell
- * Copyright (C) 2009-2012 Joel Rosdahl
+ * Copyright (C) 2009-2014 Joel Rosdahl
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -48,6 +48,11 @@ init_log(void)
 	}
 	logfile = fopen(cache_logfile, "a");
 	if (logfile) {
+		int fd = fileno(logfile);
+		int flags = fcntl(fd, F_GETFD, 0);
+		if (flags >= 0) {
+			fcntl(fd, F_SETFD, flags | FD_CLOEXEC);
+		}
 		return true;
 	} else {
 		return false;
@@ -149,10 +154,13 @@ copy_fd(int fd_in, int fd_out)
 		ssize_t count, written = 0;
 		do {
 			count = write(fd_out, buf + written, n - written);
-			if (count == -1 && errno != EINTR) {
-				fatal("Failed to copy fd");
+			if (count == -1) {
+				if (errno != EAGAIN && errno != EINTR) {
+					fatal("Failed to copy fd");
+				}
+			} else {
+				written += count;
 			}
-			written += count;
 		} while (written < n);
 	}
 
@@ -176,7 +184,7 @@ mkstemp(char *template)
 int
 copy_file(const char *src, const char *dest, int compress_dest)
 {
-	int fd_in = -1, fd_out = -1;
+	int fd_in, fd_out;
 	gzFile gz_in = NULL, gz_out = NULL;
 	char buf[10240];
 	int n, written;
@@ -684,7 +692,9 @@ dirname(char *s)
 #endif
 	if (p < p2)
 		p = p2;
-	if (p) {
+	if (p == s) {
+		return s;
+	} else if (p) {
 		*p = 0;
 		return s;
 	} else {
@@ -964,7 +974,7 @@ same_executable_name(const char *s1, const char *s2)
 
 /*
  * Compute the length of the longest directory path that is common to two
- * strings.
+ * paths. s1 is assumed to be the path to a directory.
  */
 size_t
 common_dir_prefix_length(const char *s1, const char *s2)
@@ -976,15 +986,21 @@ common_dir_prefix_length(const char *s1, const char *s2)
 		++p1;
 		++p2;
 	}
-	while (p1 > s1 && ((*p1 && *p1 != '/' ) || (*p2 && *p2 != '/'))) {
+	while ((*p1 && *p1 != '/') || (*p2 && *p2 != '/')) {
 		p1--;
 		p2--;
 	}
+	if (!*p1 && !*p2 && p2 == s2 + 1) {
+		/* Special case for s1 and s2 both being "/". */
+		return 0;
+	}
 	return p1 - s1;
 }
 
 /*
- * Compute a relative path from from to to. Caller frees.
+ * Compute a relative path from from (an absolute path to a directory) to to (a
+ * path). Assumes that both from and to are well-formed and canonical. Caller
+ * frees.
  */
 char *
 get_relative_path(const char *from, const char *to)
@@ -994,23 +1010,24 @@ get_relative_path(const char *from, const char *to)
 	const char *p;
 	char *result;
 
+	assert(from && from[0] == '/');
+	assert(to);
+
 	if (!*to || *to != '/') {
 		return x_strdup(to);
 	}
 
 	result = x_strdup("");
 	common_prefix_len = common_dir_prefix_length(from, to);
-	for (p = from + common_prefix_len; *p; p++) {
-		if (*p == '/') {
-			x_asprintf2(&result, "../%s", result);
+	if (common_prefix_len > 0 || !str_eq(from, "/")) {
+		for (p = from + common_prefix_len; *p; p++) {
+			if (*p == '/') {
+				x_asprintf2(&result, "../%s", result);
+			}
 		}
 	}
 	if (strlen(to) > common_prefix_len) {
-		p = to + common_prefix_len + 1;
-		while (*p == '/') {
-			p++;
-		}
-		x_asprintf2(&result, "%s%s", result, p);
+		x_asprintf2(&result, "%s%s", result, to + common_prefix_len + 1);
 	}
 	i = strlen(result) - 1;
 	while (i >= 0 && result[i] == '/') {
@@ -1101,7 +1118,7 @@ x_unlink(const char *path)
 	 * file. We don't care if the temp file is trashed, so it's always safe to
 	 * unlink it first.
 	 */
-	char* tmp_name = format("%s.%s.rmXXXXXX", path, tmp_string());
+	char *tmp_name = format("%s.tmp.rm.%s", path, tmp_string());
 	int result = 0;
 	cc_log("Unlink %s via %s", path, tmp_name);
 	if (x_rename(path, tmp_name) == -1) {
@@ -1124,13 +1141,6 @@ x_readlink(const char *path)
 	long maxlen = path_max(path);
 	ssize_t len;
 	char *buf;
-#ifdef PATH_MAX
-	maxlen = PATH_MAX;
-#elif defined(MAXPATHLEN)
-	maxlen = MAXPATHLEN;
-#elif defined(_PC_PATH_MAX)
-	maxlen = pathconf(path, _PC_PATH_MAX);
-#endif
 	if (maxlen < 4096) maxlen = 4096;
 
 	buf = x_malloc(maxlen);
@@ -1168,7 +1178,6 @@ read_file(const char *path, size_t size_hint, char **data, size_t *size)
 	}
 	allocated = size_hint;
 	*data = x_malloc(allocated);
-	ret = 0;
 	while (true) {
 		if (pos > allocated / 2) {
 			allocated *= 2;
